#ifndef __COMMANDER_H__
#define __COMMANDER_H__

#include <iostream>
#include <map>
#include <functional>

namespace commander
{
	class Argument;
	class Option;
	class Command;
	class CommandArgument;

	using ExecuteCallback = std::function<int(const CommandArgument &)>;

	class CommandArgument
	{
	public:
		inline CommandArgument(const std::map<std::string, std::string> &options, const std::map<std::string, std::string> &arguments) : m_options(options), m_arguments(arguments) {}

		template <typename T>
		inline T option(const std::string &name, T defaultValue) const
		{
			T res;
			getValue(m_options, res, name, defaultValue);
			return res;
		}

		template <typename T>
		inline T option(const std::string &name) const
		{
			T res;
			getValue(m_options, res, name);
			return res;
		}

		template <typename T>
		inline T argument(const std::string &name, T defaultValue) const
		{
			T res;
			getValue(m_arguments, res, name, defaultValue);
			return res;
		}

		template <typename T>
		inline T argument(const std::string &name) const
		{
			T res;
			getValue(m_arguments, res, name);
			return res;
		}

	private:
		static void getValue(const std::map<std::string, std::string> &input, bool &out, const std::string &name, bool defaultValue = false);
		static void getValue(const std::map<std::string, std::string> &input, int32_t &out, const std::string &name, int32_t defaultValue = 0);
		static void getValue(const std::map<std::string, std::string> &input, int64_t &out, const std::string &name, int64_t defaultValue = 0);
		static void getValue(const std::map<std::string, std::string> &input, float &out, const std::string &name, float defaultValue = 0);
		static void getValue(const std::map<std::string, std::string> &input, double &out, const std::string &name, double defaultValue = 0);
		static void getValue(const std::map<std::string, std::string> &input, std::string &out, const std::string &name, std::string defaultValue = "");

	private:
		std::map<std::string, std::string> m_options;
		std::map<std::string, std::string> m_arguments;
	};

	class Option
	{
	public:
		Option(const std::string &name);
		~Option();

		Option &shortcut(const std::string &shortcut);
		Option &description(const std::string &description);
		Option &required(bool required = true);
		Option &boolean(bool boolean = true);
		Option &valueName(const std::string &valueName);

	private:
		void *m_ptr;
	};

	class Argument
	{
	public:
		Argument(const std::string &name);
		~Argument();

		Argument &description(const std::string &description);
		Argument &required(bool required = true);

	private:
		void *m_ptr;
	};

	class Command
	{
	public:
		Command(const std::string &name);
		~Command();

		Command &description(const std::string &description);
		Command &command(const std::string &name);
		Command &execute(ExecuteCallback callback);
		Option &option(const std::string &name);
		Argument &argument(const std::string &name);
		std::string usage();

	private:
		void *m_ptr;
	};

	int parse(Command &command, int argc, char **argv);

}

#endif
